"use strict";

Components.utils.import("resource://gre/modules/Services.jsm");

var gDocument, gLocation, gPref, gData, gApply, gIsClosing;
var gFallbacks = 1, gSaveState = false;
var devBox;

var resettedItems;

var menuTree, attrTree;

function init(target) {
  if(!target) return;

  var containers, l, i;

  gData = []; resettedItems = []; gApply = false;
  gDocument = target.document;
  gLocation = gDocument.location.href;
  gPref = target.menumanipulator.ps;

  document.title = gStrings.title + " - " + gLocation;

  menuTree = document.getElementById("menu-tree");
  attrTree = document.getElementById("attr-tree");
  devBox = document.getElementById("devbox");

  containers = gDocument.getElementsByTagName("menubar");
  for(i = 0, l = containers.length; i < l; i++) gData = gData.concat(addData([containers[i]],0));

  containers = gDocument.getElementsByTagName("popup");
  for(i = 0, l = containers.length; i < l; i++) gData = gData.concat(addData([containers[i]],0));

  containers = gDocument.getElementsByTagName("menupopup");
  for(i = 0, l = containers.length; i < l; i++) {
   if(containers[i].parentNode.localName != "menu" && containers[i].id)
    gData = gData.concat(addData([containers[i]],0));
  }

  var tabContext = gDocument.getElementById("menumanipulator-tabcontext");
  if(tabContext) gData = gData.concat(addData([tabContext],0));

  var tabContext = gDocument.getElementById("tabContextMenu");
  if(tabContext && gDocument.getBindingParent(tabContext)) gData = gData.concat(addData([tabContext],0));

  menuTree.view = menuView;

  attrView.data = { length:0 };
  attrTree.view = attrView;

  if(gPref.getPrefType(gLocation))
    devBox.value = gPref.getCharPref(gLocation).replace(/;   /g, ";\n\n").replace(/; /g, ";\n");
  else
    devBox.value = "";

  gFallbacks = gPref.getIntPref("Move.fallbacks");
  gSaveState = gPref.getBoolPref("saveState");

  menuTree.view.selection.select(gData.length ? 0 : -1);
}


function MenuData(node, level) {
 this.node = node;
 this.level = level;

 this.label = node.getAttribute("label");

 this.Set = {};
}

MenuData.prototype = {
 toString: function() { return this.label; }
}

function addData(nodes,level) {
  var retval = [];
  for(var i = 0, l = nodes.length; i < l; i++) {
    var node = nodes[i];

    if(node.nodeType != 1) continue;

    if(node.id.indexOf("rdf:") > -1
    || node.node
    || node.id.indexOf("NC:") > -1
    || node.id.indexOf("://") > -1
    || node.hasAttribute("toolbarindex")
    || node.hasAttribute("data")
    || node.hasAttribute("index")
    ||(node.hasAttribute("image") && node.getAttribute("image").indexOf("moz-icon:") > -1) //bug 108480
    || node.className == "tooltip"
    || node.hasAttribute("userdefined")
    || node.getAttribute("menumanipulator") == "removed"
    ) continue;

    var data = new MenuData(node, level);
    if(node.menumanipulator) {
      data.path = node.menumanipulator.path;
      data.created = node.menumanipulator.created;
      data.moved = node.menumanipulator.moved;
      
      var attr;
      for(attr in node.menumanipulator.set) {
        data.Set[attr] = node.menumanipulator.set[attr];
      }
    }

    if(!data.label && (node.hasAttribute("command") || node.hasAttribute("observes"))) {
      var command = node.ownerDocument.getElementById(node.getAttribute("command") || node.getAttribute("observes"));
      if(command) data.label = command.getAttribute("label");
    }

    if(!data.label) data.label = node.id;

    if(node.hasAttribute("image")) data.image = node.getAttribute("image");
    else {
      var listStyleImage = gDocument.defaultView.getComputedStyle(node, null).listStyleImage;
      if(listStyleImage != "none" && gDocument.defaultView.getComputedStyle(node, null).MozImageRegion == "auto")
        try { data.image = listStyleImage.replace(/^url\('?"?(.*?)'?"?\)/,"$1"); } catch(err){}
    }

    retval.push(data);

    switch(node.localName) {
      case "menuitem":
        if(!data.label) retval.pop();
        break;
      case "menu":
        data.container = true; data.open = true;

        if(node.firstElementChild && node.firstElementChild.localName == "menupopup") {
         data.open = node.getAttribute("menumanipulator") != "closed";
         data.target = node.firstElementChild;
         if(data.open)
          retval = retval.concat(addData(node.firstElementChild.children, level+1));
         else
          data.children = addData(node.firstElementChild.children, level+1);
        }
        break;
      case "menuseparator":
        data.separator = true;
        break;
      case "menupopup": case "popup": case "menubar":
        data.container = true; data.open = node.getAttribute("menumanipulator") != "closed";
        data.target = node;
        if(data.open)
         retval = retval.concat(addData(node.children, level+1));
        else
         data.children = addData(node.children, level+1);
       break;
      default:
        retval.pop();
        break;
    }
  }

  return retval;
}

function createXPath(node) {
  if(node.id) return "id('"+specEscape(node.id)+"')";
  if(node.nodeType == 9) return "";

  var path = "x:" + node.localName;
  if(node.hasAttribute("command")) path += "[@command='" + specEscape(node.getAttribute("command")) +"']";
  else if(node.hasAttribute("label")) path += "[@label='" + specEscape(node.getAttribute("label")) +"']";
  else if(node.hasAttribute("observes")) path += "[@observes='" + specEscape(node.getAttribute("observes")) +"']";

  return createXPath(node.parentNode) + "/" + path;
}

var menuView = {
  get rowCount() { return gData.length; },

  getColumnProperties: function(){},

  isContainer: function(index) { return gData[index].container; },
  isContainerOpen: function(index) { return gData[index].open; },
  isContainerEmpty: function(index) { return !(gData[index].target); },
  toggleOpenState: function(index) {
   var currentRow = gData[index];
   if(currentRow.open) {
    var length = 0;
    while(gData[index + ++length] &&
          gData[index].level < gData[index + length].level);
    if(length-- > 0) {
     currentRow.children = gData.splice(index + 1, length);
     menuView.treebox.rowCountChanged(index + 1,-length);
    }
    currentRow.node.setAttribute("menumanipulator","closed");
   } else {
    for(var i = 0, l = currentRow.children.length; i < l; i++) {
     gData.splice(index + 1 + i, 0, currentRow.children[i]);
    }
    menuView.treebox.rowCountChanged(index + 1, currentRow.children.length);
    delete currentRow.children;
    currentRow.node.removeAttribute("menumanipulator");
   }

   gData[index].open = !gData[index].open;
   this.treebox.invalidateRow(index);

  },

  isSeparator: function(index) { return gData[index].separator; },
  isSorted: function() { return false; },

  canDrop: function(index, orientation, dataTransfer) {
    if(dataTransfer.mozSourceNode != this.treebox.treeBody ||
       index == this.selection.currentIndex ||
       (orientation == 0 && !gData[index].target) ||
       (index == 0 && orientation == -1)
      )
    return false;

    var parentIndex = this.getParentIndex(index);
    while(parentIndex != -1) {
      if(this.selection.currentIndex == parentIndex) return false;
      parentIndex = this.getParentIndex(parentIndex);
    }

    return true;
  },
  onDragStart: function (event)
  {
    var item = gData[menuTree.currentIndex];
    if(item.level > 0 &&
       (item.node.localName != "menuseparator" || item.node.id) &&
       (item.path || createXPath(item.node).split("/").length < 4)
      ) {
      event.dataTransfer.setData("text/x-menumanipulator", "")
    }
  },
  drop: function(row, orientation) {
    var length = 0;
    while(gData[menuTree.currentIndex + ++length] &&
          gData[menuTree.currentIndex].level < gData[menuTree.currentIndex + length].level);

    var targetRow = gData[row];
    var newIndex, levelAdjust = targetRow.level - gData[menuTree.currentIndex].level;
    if(orientation == -1) {
      newIndex = row;

      if(targetRow.level == 0) levelAdjust++;
    } else if(orientation == 0) {
      var lastChild = row;
      while(gData[++lastChild] && gData[lastChild].level > targetRow.level);
      newIndex = lastChild;

      levelAdjust++;
    } else {
      newIndex = row + 1;

      if(targetRow.target && targetRow.open) levelAdjust++;
    }

    if(newIndex > menuTree.currentIndex) newIndex -= length;
    if(newIndex == menuTree.currentIndex && levelAdjust == 0) return;

    var currentRow = gData[menuTree.currentIndex];
    gApply = true;
    currentRow.move = true;
    currentRow.path = createXPath(currentRow.node);

    function adjustLevel(data, level) {
      var i, l;
      for(i = 0, l = data.length; i < l; i++) {
        data[i].level += level;
        if(data[i].children) adjustLevel(data[i].children, level);
      }
    }

    var movedData = gData.splice(menuTree.currentIndex,length);
    adjustLevel(movedData, levelAdjust);

    var newContainer = gData;
    if(targetRow.children && orientation == 0) {
      newContainer = targetRow.children;
      newIndex = targetRow.children.length;
      currentRow = null;
      menuView.treebox.rowCountChanged(menuTree.currentIndex,-length);
    }

    var i, l;
    for(i = 0, l = movedData.length; i < l; i++)
      newContainer.splice(newIndex + i, 0, movedData[i]);

    this.treebox.invalidate();

    if(currentRow) {
      i = -1;
      do { i++; } while(currentRow != gData[i]);
      this.selection.select(i);
      this.treebox.ensureRowIsVisible(i);
    }

  },

  getLevel: function(index) { return gData[index].level; },

  getImageSrc: function(row, col) { return gData[row].image; },

  getCellText : function(row,col){ return gData[row].label;},

  getRowProperties: function(row) { this.getCellProperties(row, ""); },
  getCellProperties: function(row, col) {
    var properties = [];

    properties.push("FilenameColumn");
    if(gData[row].container) properties.push("directory");
    if(gData[row].separator) properties.push("separator");

    if(gData[row].created) properties.push("created");
    if(gData[row].move || gData[row].moved) properties.push("moved");

    var SetLength = Object.keys(gData[row].Set).length;
    if (SetLength > 1 || (SetLength == 1 && !gData[row].Set["menumanipulator"]))
      properties.push("modified");

    return properties.join(" ");
  },

  setTree: function(treebox) { this.treebox=treebox; },

  cycleHeader: function cycleHeader(col, elem) {  },

  hasNextSibling: function(rowIndex, afterIndex) {
    while(gData[++afterIndex] && gData[afterIndex].level >= gData[rowIndex].level)
      if(gData[afterIndex].level == gData[rowIndex].level) return afterIndex;

    return 0;
  },

  getParentIndex: function(rowIndex) { 
    var i = rowIndex;

    while(gData[--i]) if(gData[i].level < gData[rowIndex].level) return i;

    return -1;
  },

  selectionChanged: function() {
    attrTree.treeBoxObject.rowCountChanged(0, -attrView.rowCount);
    attrView.data = [];

    if(menuTree.currentIndex == -1) return;

    var item = gData[menuTree.currentIndex];

    var i, l;
    for(i = 0, l = item.node.attributes.length; i < l; i++) {
      if(item.node.attributes[i].name in item.Set) continue;
      attrView.data.push({attrName: item.node.attributes[i].name, attrValue: item.node.attributes[i].value});
    }
    var attr;
    for(attr in item.Set)
      attrView.data.push({attrName: attr, attrValue: item.Set[attr]});

    attrView.data.sort(function(a,b){ return a.attrName.localeCompare(b.attrName); });

    attrTree.treeBoxObject.rowCountChanged(0, attrView.rowCount);
  }

}

function onOK() {
  var i = -1, l, item;

  while(gData[++i]) {
   item = gData[i];
   item.node.removeAttribute("menumanipulator");
   delete item.Set["menumanipulator"];
   delete item.node.menumanipulator;

   if(item.container && !item.open) {
    menuView.toggleOpenState(i);
    item.node.setAttribute("menumanipulator","closed");
    if(gSaveState) { 
     item.Set["menumanipulator"] = "closed";
    }
   }
   if(item.created || item.moved || item.move || item.Set.toSource().length > 4)
    item.save = true;
  }

  var order = [], level;
  for(level = 0; level < 10; level++)
    for(i = gData.length - 1; i > -1 ; i--)
      if(gData[i].level == level && gData[i].save)
        order.push(i);

  var actions = [];
  for(i = 0, l = order.length; i < l; i++) actions.push(generateActions(gData[order[i]],order[i]));

  for(i = 0, l = resettedItems.length; i < l; i++) resettedItems[i].node.setAttribute("menumanipulator","removed");

  var str = Components.classes["@mozilla.org/supports-string;1"].createInstance(Components.interfaces.nsISupportsString);
  str.data = actions.join("  ");
  gPref.setComplexValue(gLocation, Components.interfaces.nsISupportsString, str);
}

function generateActions(item, row) {
  var retval = "";

  if(!item.path) item.path = createXPath(item.node);

  item.node.menumanipulator = { path: item.path, set: item.Set };

  if(!item.created)
    retval += 'Node("'+ item.path +'"); ';

  if(item.moved || item.move) {
    item.node.menumanipulator.moved = true;
    var nextSibling = menuView.hasNextSibling(row,row);
    var relNode = nextSibling ? gData[nextSibling].node : gData[menuView.getParentIndex(row)].target;

    if(item.move) {
      if(nextSibling)
        relNode.parentNode.insertBefore(item.node,relNode);
      else {
        relNode.appendChild(item.node);
      }
    }

    retval += 'Move("'+ createXPath(relNode) +'",'+ (nextSibling ? -1 : -2);

    if(nextSibling) {
      var i, l, fallbacks = [];
      for(i = 0, l = gFallbacks; i < l; i++) {
        nextSibling = menuView.hasNextSibling(row,nextSibling);
        if(!nextSibling) { fallbacks.push('"' + createXPath(gData[menuView.getParentIndex(row)].target) + '",-2'); break; }
        fallbacks.push('"' + createXPath(gData[nextSibling].node) + '",-1')
      }
      if(fallbacks) retval += ",[" + fallbacks + "]";
    }

    retval += '); ';
  }
    
  var attr;

  for(attr in item.Set) {
    item.node.setAttribute(attr, item.Set[attr]);
    retval += "Set('"+specEscape(attr)+"','"+specEscape(item.Set[attr])+"'); ";
  }

  if(item.created) {
    item.node.menumanipulator.created = true;
    retval = "New('"+item.node.localName+"','"+specEscape(item.node.id)+"'); " + retval;
    if(item.node.localName == "menu")
      retval+= "New('menupopup'); Move(\"id('" + specEscape(item.node.id) + "')\",-2); ";
  }

  return retval;
}

function specEscape(val) {
 return val.replace(/\\/g,"\\\\").replace(/'/g,"\\'").replace(/\n/g,"\\n").replace(/\r/g,"");
}

function devSave() {
  gPref.setCharPref(gLocation, devBox.value);
}

var attrView = {
  setTree: function(treebox) { this.treebox=treebox; },

  get rowCount() { return this.data.length; },

  getCellText: function(row,col){ return this.data[row][col.id || col]},

  isContainer: function() { return false; },
  isSeparator: function() { return false; },
  isSorted: function() { return false; },

  getColumnProperties: function() {},
  getRowProperties: function(row, prop) { this.getCellProperties(row, ""); },
  getCellProperties: function(row, col) {
    var properties = [];

    if(attrView.data[row].attrName != "menumanipulator" &&
       gData[menuTree.currentIndex].Set[attrView.data[row].attrName]) properties.push("modified");
    
    return properties.join(" ");
  },
  getImageSrc: function() { return ""; },

  cycleHeader: function() {},

  onEdit: function() {
    if(attrTree.currentIndex == -1) return;

    var attr = this.data[this.selection.currentIndex];
    if(!attr) return;

    var input = {name: attr.attrName, value: attr.attrValue};
    openDialog("edit.xul", "_blank", "resizable,modal,centerscreen",gStrings.editTitle,input);

    if(attr.attrValue == input.value) return;
    attr.attrValue = input.value;
    this.treebox.invalidate();

    var item = gData[menuTree.currentIndex];
    if(!item.path) item.path = createXPath(item.node);
    gApply = true;
    item.Set[attr.attrName] = attr.attrValue;

    if(attr.attrName == "label")
     item.label = attr.attrValue;
    menuView.treebox.invalidateRow(menuTree.currentIndex);
  },

  onInsert: function() {
    if(menuTree.currentIndex == -1) return;

    var attr = {}
    var input = {value: "", new: true};

    Services.prompt.prompt(window, gStrings.newTitle, gStrings.newText, input, null, {});
    if(!input.value) return;

    attr.attrName = input.name = input.value;
    input.value = null;

    openDialog("edit.xul", "_blank", "resizable,modal,centerscreen",gStrings.newTitle,input);

    if(attr.attrName && input.accepted) {
      attr.attrValue = input.value;
      this.data.push(attr);
      this.treebox.rowCountChanged(this.data.length - 1, 1);
      var item = gData[menuTree.currentIndex];
      if(!item.path) item.path = createXPath(item.node);
      gApply = true;
      item.Set[attr.attrName] = attr.attrValue;
      menuView.treebox.invalidateRow(menuTree.currentIndex);
    }

  },

  onReset: function() {
    if(attrTree.currentIndex == -1) return;

    var attr = this.data[this.selection.currentIndex];
    var item = gData[menuTree.currentIndex];
    delete item.Set[attr.attrName];

    if(attr.attrName == "label" && item.node.attributes["label"])
     item.label = item.node.attributes["label"].value;

    attr.attrValue = gStrings.attrOnReset;
    this.treebox.invalidate();
    menuView.treebox.invalidateRow(menuTree.currentIndex);

    gApply = true;
  }
}

function createNew(type) {
  if(gData.length == 0) return;

  var currentIndex = menuTree.currentIndex == -1 ? 0 : menuTree.currentIndex;

  var item = {
    node: gDocument.createElement(type), Set: new Object(), move: true, created: true
  }

  item.node.id = type+Date.now();

  switch(type) {
    case "menu":
      item.container = true; item.open = true;
      item.target = item.node.appendChild(gDocument.createElement("menupopup"));
      if(gData[currentIndex].node.localName == "menubar") item.Set["style"] = "";
    case "menuitem":
      var input = {value: ""}
      if(!Services.prompt.prompt(window, gStrings.newItem.replace("$1", type), gStrings.newItemText, input, null, {})) return;

      var label = input.value;

      item.label = label;
      item.Set["label"] = label;
      item.node.setAttribute("label",label);
      break;
    case "menuseparator":
      item.label = item.node.id;
      item.separator = true;
      break
  }

  if(gData[currentIndex].target && !gData[currentIndex].open) menuView.toggleOpenState(currentIndex);
  item.level = gData[currentIndex].target ? gData[currentIndex].level + 1 : gData[currentIndex].level;

  var newIndex = currentIndex + 1;
  gData.splice(newIndex,0,item);
  menuTree.treeBoxObject.rowCountChanged(newIndex,1);
  menuView.selection.select(newIndex);

  gApply = true;
}

function Reset(all) {
  var ok;
  if(all) {
    ok = confirm(gStrings.promptresetall);
    if(ok) {
      gPref.clearUserPref(gLocation);
      window.close();
    }
  } else {
    if(menuTree.currentIndex > -1) {
      gApply = true;
      var length = 0;
      while(gData[menuTree.currentIndex + ++length] &&
            gData[menuTree.currentIndex].level < gData[menuTree.currentIndex + length].level);

      resettedItems.push(gData[menuTree.currentIndex]);

      gData.splice(menuTree.currentIndex,length);
      menuView.treebox.rowCountChanged(menuTree.currentIndex,-length);
    }
  }
}

function switchWindow(event)
{
  if(gApply && !confirm(gStrings.onCancel)) return;

  var mediator = Components.classes["@mozilla.org/rdf/datasource;1?name=window-mediator"]
                 .getService(Components.interfaces.nsIWindowDataSource);

  var target = mediator.getWindowForResource(event.target.getAttribute('id'));

  if (target) init(target);
}

function onCancel() {
  if(gApply && gIsClosing) {
    gIsClosing = false;
    return confirm(gStrings.onCancel);
  }
  return true;
}
